"""Class definition for GeoModeller XML-Files
This version includes drillholes
Specific methods are defined for the uncertainty analysis (in combination
with Uncertainty_Obj module)

(c) J. Florian Wellmann, 2009-2013
"""

try:
    import elementtree.ElementTree as ET
except ImportError:
    try:
        import etree.ElementTree as ET
    except ImportError:
        try:
            import xml.etree.ElementTree as ET
        except ImportError:
            import ElementTree as ET
# import Latex_output_5 as LO
from pylab import *
import string

# python module to wrap GeoModeller XML file and perform all kinds of data
# procedures, e.g.:
# - Stochastic data modeling
# - TWT to depth conversion
# - Documentation module
#
# To Do:
# - exception handling and passing
# - log file?
# - check validity of xml-code, test, if past-processing with sax neccessary?
# - implement auto-documentation
# - clear-up spaghetti code!!!!! Check dependencies and other modules
#    to get a consistent lay-out

class GeomodellerClass:
    """Wrapper for GeoModeller XML-datafiles to perform all kinds of data
    manipulation and analysis on low level basis, e.g.:
    - Uncertainty Simulation
    - TWT to depth conversion
    - Data analysis, e.g. Auto-documentation"""
    def load_geomodeller_file(self, xml_file):
        self.xml_file_name = xml_file
        try:
            tree = ET.parse(xml_file)
        except IOError, (nr, string_err):
            print "Can not open xml File " + xml_file + ": " + string_err
            print "Please check file name and directory and try again"
            raise IOError
        # safe tree on local varibale
        self.tree = tree
        # get rootelement
        self.rootelement = tree.getroot()
        # set other class variables
        self.xmlns = "http://www.geomodeller.com/geo"
        self.gml = "http://www.opengis.net/gml"
        
    def load_deepcopy_tree(self, deepcopy_tree):
        """load tree information from deepcopied tree into object"""
        self.tree = deepcopy_tree
        self.rootelement = deepcopy_tree.getroot()
        # set other class variables
        self.xmlns = "http://www.geomodeller.com/geo"
        self.gml = "http://www.opengis.net/gml"
        
    def deepcopy_tree(self):
        """create a deep copy of original tree to restore later, e.g. for uncertainty evaluation"""
            
        deepcopy_tree = deepcopy(self.tree)
        deepcopy_tree.parent = None
        return deepcopy_tree
        
        
    def reload_geomodeller_file(self, deepcopy_tree):
        """restore original tree root from deep copy of orignial tree
        deep copy can be created (not automatically to save memory!) with
        self.deepcopy_tree()
        """
        try:
            self.tree = deepcopy_tree
            self.rootelement = tree.getroot()
        except NameError:
            print "No deep copy of original tree available, please create with self.deepcopy_tree()"
            
    def get_sections(self):
        """get sections out of rootelement, safe array with section elements 
        in local variable"""
        sections_parent = self.rootelement.findall("{"+self.xmlns+"}Sections")[0]
        self.sections = sections_parent.findall("{"+self.xmlns+"}Section")
        return self.sections
    
    def get_faults(self):
        """get fault elements out of rootelement and safe as local list"""
        try:
            faults_parent = self.rootelement.findall("{"+self.xmlns+"}Faults")[0]
            self.faults = faults_parent.findall("{"+self.xmlns+"}Fault")
        except IndexError:
            print("No faults found in model")
    
    def get_formations(self):
        """get formation elements out of rootelement and safe as local list"""
        formations_parent = self.rootelement.findall("{"+self.xmlns+"}Formations")[0]
        self.formations = formations_parent.findall("{"+self.xmlns+"}Formation")
    
    def get_stratigraphy_list(self,**kwds):
        """get project stratigraphy and return as list; lowermost formation: 1
        for GeoModeller dll access (this ist the formation number that is returned with
        the GetComputedLithologyXYZ function in the geomodeller dll
        optional keywords:
        out = string : set 'out' formation to this name (might be necessary for TOUGH2 simulation!)
        """
        series_list = []
        strati_column = self.rootelement.find("{"+self.xmlns+"}GeologicalModel").find("{"+self.xmlns+"}ProjectStratigraphicColumn")#.findall("{"+self.xmlns+"Series")
        series = strati_column.findall("{"+self.xmlns+"}Series")
        for s in series:
            data = s.findall("{"+self.xmlns+"}Data")
            for d in data:
                series_list.append(d.get("Name"))
        # append "out" as uppermost formation for "out values
        if kwds.has_key('out'):
            series_list.append(kwds['out'])
        else:
            series_list.append("out")
        self.stratigraphy_list = series_list
        return series_list
    
    def get_provenances(self):
        """get provenance table and return as dictionary with provenance rank as key
        deprecated, use get_provenance_table() instead!
        """
        print "deprecated, use get_provenance_table() instead!"
        provenance_parent = self.rootelement.find("{"+self.xmlns+"}ProvenanceTable")
        rows = provenance_parent.find("{"+self.xmlns+"}Result").findall("{"+self.xmlns+"}Row")
        self.prov_dict = {}
        for row in rows:
            line = row.text.split(",")
            #check for leading and ending '"' signs
            if line[0][0] == '"':
                self.prov_dict[line[0][1:-1]] = line[1][1:-1]
            else:
                self.prov_dict[line[0]] = line[1]

    def get_provenance_table(self):
        """get provenance table and return as dictionary with provenance rank as key"""
        provenance_parent = self.rootelement.find("{"+self.xmlns+"}ProvenanceTable")
        rows = provenance_parent.find("{"+self.xmlns+"}Result").findall("{"+self.xmlns+"}Row")
        self.prov_dict = {}
        for row in rows:
            line = row.text.split(",")
            #check for leading and ending '"' signs
            if line[0][0] == '"':
                self.prov_dict[line[0][1:-1]] = line[1][1:-1]
            else:
                self.prov_dict[line[0]] = line[1]
    
    def set_provenance_table(self, provenance_dict):
        """create provenance table from dictionary; can be used to extend existing
        provenance table as a first step to assign uncertainty values
        attention: old provenance table is deleted!
        """
        provenance_parent = self.rootelement.find("{"+self.xmlns+"}ProvenanceTable")
        result = provenance_parent.find("{"+self.xmlns+"}Result")
        rows = result.findall("{"+self.xmlns+"}Row")
        # delete old provenance table
        for row in rows:
            result.remove(row)
        # define new table from provenance dictionary
        for l in provenance_dict:
            elem = ET.Element("{"+self.xmlns+"}Row")
            # turn list into string
            text = '"%s","%s"' % (l,procenance_dict[l])
            elem.text = text
            results.append(elem)
        
    def get_provenance(self, element):
        """get provenance of an element, return as string
        if not defined: returns None"""
        return element.get("Provenance")
        
    def get_section_names(self):
        """get all section names out of local variable self.sections"""
        # todo: return section names as dictionary with element and name?
        # test if self.sections defined, if not -> create
        try:
            self.sections
        except AttributeError:
            # print "Create sections Data array"
            self.get_sections()
        section_names = []
        for section in self.sections:
            section_names.append(section.get("Name"))
        return section_names
    
    def get_formation_names(self):
        """get formation names and return as list"""
        forms=[]
        try:
            self.formations
        except AttributeError:
            self.get_formations()
        for formation in self.formations:
            forms.append(formation.get("Name"))
        return forms

    def get_points_in_sections(self):
        """Create dictionary of all points (with obs-id) in all sections"""
        self.create_sections_dict()
        for sec in self.section_dict.keys():
            forms = self.get_formation_point_data(self.section_dict[sec])
            if forms == None:
                print "\t\t\tNo Formation Points in this section"
            else:
                for form in forms:
 
                    data = form.find("{"+self.xmlns+"}Data")
                    print "\nObsID = %s" % form.get("ObservationID")
                    print "\tFormation name\t= %s" % data.get("Name")
                    element_point = form.find("{"+self.gml+"}LineString")
                    element_coords = element_point.find("{"+self.gml+"}coordinates")
                    tmp = element_coords.text.split(" ")
                    for tmp1 in tmp:
                        if tmp1 == '': continue
                        tmp_cds = tmp1.split(",")
                        print("\tX = %.1f, Y = %.1f" % (float(tmp_cds[0]), float(tmp_cds[1])))


                    fol = form.find("{"+self.xmlns+"}FoliationObservation")
                    if fol is not None:
                        print("\tFoliation defined: azimuth = %.1f, dip = %.1f" % (float(fol.get("Azimuth")), float(fol.get("Dip"))))
                        # get position of foliation (yet another point)
                        pt = fol.find("{"+self.gml+"}Point")
                        c = pt.find("{"+self.gml+"}coordinates")
                        cds = c.text.split(",")
                        print("\t\tX = %.1f, Y = %.1f" % (float(cds[0]), float(cds[1])))

            print "\n"
            print 80*"-"
            print "Foliations in section %s:" % sec
            print 80*"-"
            foliations = self.get_foliations(self.section_dict[sec])
            if foliations == None:
                print "\t\t\tNo foliations in this section"
            else:
                for fol1 in foliations:
                    print "\nObsID = %s" % fol1.get("ObservationID")
                    data = fol1.find("{"+self.xmlns+"}Data")
                    fol = fol1.find("{"+self.xmlns+"}FoliationObservation")
                    print "\tFormation name\t= %s" % data.get("Name")
                    print("\tAzimuth = %.1f, dip = %.1f" % (float(fol.get("Azimuth")), float(fol.get("Dip"))))
                    pt = fol.find("{"+self.gml+"}Point")
                    c = pt.find("{"+self.gml+"}coordinates")
                    cds = c.text.split(",")
                    print("\tX = %.1f, Y = %.1f" % (float(cds[0]), float(cds[1])))

    
    def get_formation_parameters(self):
        """read formation parameters; physical
        properties, density, th. cond etc... store in dict"""
        #
        # To do: re-write in a more elegant way and keep original 
        # structure and key-words?
        # 
        self.formation_params = {}
        try:
            self.formations
        except AttributeError:
#            print "Create sections Data array"
            self.get_formations()
        for formation in self.formations:
            self.formation_params[formation.get("Name")] = {}
            geophys = formation.find("{"+self.xmlns+"}Geophysics")
            dens = geophys.find("{"+self.xmlns+"}DensityCompoundDistribution")
            dens_simple = dens.find("{"+self.xmlns+"}SimpleDistribution")
            self.formation_params[formation.get("Name")]["dens_mean"] = dens_simple.get("Mean")
            self.formation_params[formation.get("Name")]["dens_law"] = dens_simple.get("LawType")
            self.formation_params[formation.get("Name")]["dens_dev"] = dens_simple.get("Deviation")
#             print geophys.getchildren()
            mag = geophys.find("{"+self.xmlns+"}RemanantMagnetizationCompoundDistribution")
            mag_simple = mag.find("{"+self.xmlns+"}SimpleDistributionVector")
            self.formation_params[formation.get("Name")]["mag"] = mag_simple.get("Mean")
            velocity = geophys.find("{"+self.xmlns+"}VelocityCompoundDistribution")
            velocity_simple = velocity.find("{"+self.xmlns+"}SimpleDistribution")
            self.formation_params[formation.get("Name")]["vel_mean"] = velocity_simple.get("Mean")
            self.formation_params[formation.get("Name")]["vel_law"] = velocity_simple.get("LawType")
            self.formation_params[formation.get("Name")]["vel_dev"] = velocity_simple.get("Deviation")
            # Thermal properties are only defined in newer versions of GeoModeller! thus check!
            
            th_cond = geophys.find("{"+self.xmlns+"}ThermalConductivityCompoundDistribution")
            if th_cond == None: continue
            th_cond_simple = th_cond.find("{"+self.xmlns+"}SimpleDistribution")
            self.formation_params[formation.get("Name")]["th_cond_mean"] = th_cond_simple.get("Mean")
            self.formation_params[formation.get("Name")]["th_cond_law"] = th_cond_simple.get("LawType")
            self.formation_params[formation.get("Name")]["th_cond_dev"] = th_cond_simple.get("Deviation")
            heat_prod = geophys.find("{"+self.xmlns+"}HeatProductionRateCompoundDistribution")
            heat_prod_simple = heat_prod.find("{"+self.xmlns+"}SimpleDistribution")
            self.formation_params[formation.get("Name")]["heat_prod_mean"] = heat_prod_simple.get("Mean")
            self.formation_params[formation.get("Name")]["heat_prod_law"] = heat_prod_simple.get("LawType")
            self.formation_params[formation.get("Name")]["heat_prod_dev"] = heat_prod_simple.get("Deviation")
    
            # same for other properties
#            print th_cond
            #
            # !!! only simple distributions yet impl.
            #
            
    def get_drillhole_elements(self):
        """get drillhole elements and store in dictionary"""
        try:
            drillhole_parent = self.rootelement.find("{"+self.xmlns+"}DrillHoles").find("{"+self.xmlns+"}GeneralDrillholes")
        except AttributeError:
            print "Problem with drillhole element; check if drillholes are defined in project!"
            return
        self.drillholes = {}
        self.drillholes["geology"]= drillhole_parent.find("{"+self.xmlns+"}GeologyTable")
        self.drillholes["collar"] = drillhole_parent.find("{"+self.xmlns+"}CollarTable")
        self.drillholes["survey"] = drillhole_parent.find("{"+self.xmlns+"}SurveyTable")
        return True 
            
    def get_drillholes_old(self):
        """get drillhole tables as elements"""
        drillhole_parent = self.rootelement.find("{"+self.xmlns+"}DrillHoles").find("{"+self.xmlns+"}GeneralDrillholes")
        ct = {}
        st = {}
        gt = {}
        ct["parent"] = drillhole_parent.find("{"+self.xmlns+"}CollarTable")
        st["parent"] = drillhole_parent.find("{"+self.xmlns+"}SurveyTable")        
        gt["parent"] = drillhole_parent.find("{"+self.xmlns+"}GeologyTable")
        ct["head"] = ct["parent"].find("{"+self.xmlns+"}Head")
        ct["result"] = ct["parent"].find("{"+self.xmlns+"}Result")        
        st["head"] = st["parent"].find("{"+self.xmlns+"}Head")
        st["result"] = st["parent"].find("{"+self.xmlns+"}Result")        
        gt["head"] = gt["parent"].find("{"+self.xmlns+"}Head")
        gt["result"] = gt["parent"].find("{"+self.xmlns+"}Result")       
        # print ct
        # set as global arguments
        self.collar = ct
        self.geology = gt
        self.survey = st
        return True
    
    def append_drillhole_data(self, element, data_list):
        """append data in list as drillhole data (i.e. new Row elements); 
        element should
        be one of the drillhole file elements, i.e. 
        self.drillholes["survey"], self.drillholes["collar"] 
        or self.drillholes["geology"] 
        list should be on correct format"""
        results = element.find("{"+self.xmlns+"}Result")
        # rows = results.findall("{"+self.xmlns+"}Row")
        # data = []
        elem = ET.Element("Row")
        # turn list into string
        text = ""
        for l in data_list:
            text = text + '%s, ' % l
        # delete last comma delimiter
        text = text[0:-2]
        elem.text = text
        results.append(elem)
        return True

    def delete_drillhole_data(self, element):
        """delete all drillhole data of an element (i.e. delete
        all Row elements; element should be one of
        self.drillholes["survey"], self.drillholes["collar"] 
        or self.drillholes["geology"] 
        """
        results = element.find("{"+self.xmlns+"}Result")
        rows = results.findall("{"+self.xmlns+"}Row")
        for row in rows:
            results.remove(row)
        
    def set_drillhole_data(self, element, data_list):
        """set data in list as drillhole data (i.e. delete all
        existing Row elements and create new ones); 
        element should
        be one of the drillhole file elements, i.e. 
        self.drillholes["survey"], self.drillholes["collar"] 
        or self.drillholes["geology"] 
        list should be on correct format"""
        results = element.find("{"+self.xmlns+"}Result")
        rows = results.findall("{"+self.xmlns+"}Row")
        for row in rows:
            results.remove(row)
        # data = []
        if element.tag == "{"+self.xmlns+"}GeologyTable": 
            for l in data_list:
                elem = ET.Element("{"+self.xmlns+"}Row")
                # turn list into string
                text = '"%s","%s","%s","%s"' % (l[0], l[1], l[2], l[3])
                elem.text = text
                results.append(elem)
        if element.tag == "{"+self.xmlns+"}CollarTable": 
            for l in data_list:
                elem = ET.Element("{"+self.xmlns+"}Row")
                # turn list into string
                text = '"%s","%s","%s","%s","%s"' % (l[0], l[1], l[2], l[3], l[4])
                elem.text = text
                results.append(elem)
        if element.tag == "{"+self.xmlns+"}SurveyTable":
            for l in data_list:
                elem = ET.Element("{"+self.xmlns+"}Row")
                # turn list into string
                text = '"%s","%s","%s","%s"' % (l[0], l[1], l[2], l[3])
                elem.text = text
                results.append(elem)
        return True
            
    def get_drillhole_data(self, element):
        """get drillhole data from result element as list
        element should be one of the drillhole file elements, i.e.
        self.drillholes["survey"], self.drillholes["collar"] 
        or self.drillholes["geology"] 
        also performs some type conversion, etc. based on type
        of element (e.g. in GeologyTable: set from and to as float values!)
        and removes leading and tailing '"'
        """
        results = element.find("{"+self.xmlns+"}Result")
        rows = results.findall("{"+self.xmlns+"}Row")
        data = []
        for row in rows:
            r = row.text.split(",")
            # now, some conversion and formatting issues:
            # remove " " around entries
            for i,l in enumerate(r):
                r[i] = l[1:-1]
                # set 1,2 to float for geology table
                if element.tag == "{"+self.xmlns+"}GeologyTable":
                    if i == 1 or i == 2:
                        r[i] = float(r[i])
                # set col 1,2,3 to float for survey table
                if element.tag == "{"+self.xmlns+"}SurveyTable":
                    if i == 1 or i == 2 or i == 3:
                        r[i] = float(r[i])
                # set col 1,2,3,4 to float for collar table
                # check: set x,y,z as int???
                if element.tag == "{"+self.xmlns+"}CollarTable":
                    if i == 1 or i == 2 or i == 3 or i == 4:
                        r[i] = float(r[i])                
            data.append(r)
        return data
    
#    def set_drillhole_data(self, element, l):
#        """set drillhole data for element from list
#        element should be one of the drillhole file elements, i.e.
#        self.drillholes["survey"], self.drillholes["collar"] 
#        or self.drillholes["geology"] """
#        return True
#     
    def create_fault_dict(self):
        """create dictionary for fault elements with names as keys"""
        # test if self.formations defined, if not -> create
        try:
            self.faults
        except AttributeError:
            print "Create Formations list"
            self.get_faults()
        self.fault_dict = {}
        for fault in self.faults:
            self.fault_dict[fault.get("Name")] = fault
        return self.fault_dict
    
    def create_formation_dict(self):
        """create dictionary for formation elements with formation names as keys"""
        # test if self.formations defined, if not -> create
        try:
            self.formations
        except AttributeError:
            print "Create formation dictionary"
            self.get_formations()
        self.formation_dict = {}
        for formation in self.formations:
            self.formation_dict[formation.get("Name")] = formation
        return self.formation_dict
          
    def create_sections_dict(self):
        """create dictionary for section elements with section names as keys
        (for easier use...)"""
        # test if self.sections defined, if not -> create
        try:
            self.sections
        except AttributeError:
            # print "Create sections dictionary"
            self.get_sections()
        self.section_dict = {}
        for section in self.sections:
            self.section_dict[section.get("Name")] = section
        return self.section_dict
            
    def get_fault_parameters(self):
        """get fault parameters out of """
        pass
    
    def get_foliations(self, section_element):
        """get all foliation data elements from a for section"""
        tmp_element = section_element.find("{"+self.xmlns+"}Structural2DData")
        # check in case there is no foliation defined in this section
        # tmp_element2 = tmp_element.find("{"+self.xmlns+"}Foliations")
        try:
            tmp_element2 = tmp_element.find("{"+self.xmlns+"}Foliations")
        except AttributeError:
            return None
        try:
            foliations = tmp_element2.findall("{"+self.xmlns+"}Foliation") 
        except AttributeError:
            return None
        return foliations
    
    def get_foliation_dip(self, foliation_element):
        """get dip of foliation element"""
        return float(foliation_element.find("{"+self.xmlns+"}FoliationObservation").get("Dip"))

    def get_foliation_azimuth(self, foliation_element):
        """get dip of foliation element"""
        return float(foliation_element.find("{"+self.xmlns+"}FoliationObservation").get("Azimuth"))
    
    def get_folation_polarity(self, foliation_element):
        """get polarity of foliation element; return true if Normal Polarity"""
        return foliation_element.find("{"+self.xmlns+"}FoliationObservation").get("NormalPolarity")

    def get_formation_data(self, section_element):
        """not used any more! use get_formation_point_data(section_element) instead"""
        print "not used any more! use get_formation_point_data(section_element) instead"
        return None

    def get_formation_point_data(self, section_element):
        """get all formation point data elements from a for section"""
        tmp_element = section_element.find("{"+self.xmlns+"}Structural2DData")
        # check in case there is no formation points defined in this section
        try:
            tmp_element2 = tmp_element.find("{"+self.xmlns+"}Interfaces")
        except AttributeError:
            return None
        return tmp_element2.findall("{"+self.xmlns+"}Interface")
    
    def get_name(self, section_element):
        """get the name of any section element (if defined)"""
        return section_element.find("{"+self.xmlns+"}Name")
    
    def get_interface_name(self, interface_element):
        """get name of interface, i.e. the formation"""
        return interface_element.find("{"+self.xmlns+"}Data").get("Name")
    
    def change_formation_point_pos(self, element, **args):
        """change position of formation point in section element
        arguments:
        x_coord, y_coord : set to this coordinates
        add_x_coord, add_y_coord : add values to existing coordinates
        use if_name = and if_provenance = to add conditions!"""
        prov = element.get("Provenance")
        name = element.find("{"+self.xmlns+"}Data").get("Name")
        if args.has_key("if_name"):
            if args["if_name"] != name: return
        if args.has_key("if_provenance"):
            if args["if_provenance"] != prov: return
        # element_fol = element.find("{"+self.xmlns+"}")
        element_point = element.find("{"+self.gml+"}LineString")
        element_coords = element_point.find("{"+self.gml+"}coordinates")
        point_list = element_coords.text.split(" ")
        if point_list[-1] == '':
            point_list = point_list[0:-1]
        if len(point_list) > 1:
            x_coords = []
            y_coords = []
            print point_list
            for point in point_list:
                # if point == '': continue
                a = point.split(',')
                [x_coord, y_coord] = [float(a[0]), float(a[1])]
                x_coords.append(x_coord)
                y_coords.append(y_coord)
            # convert to arrays for calculation
            x_coords = array(x_coords)
            y_coords = array(y_coords)
            if args.has_key("add_x_coord"):
                x_coords = x_coords + float(args["add_x_coord"])
            if args.has_key("add_y_coord"):
                y_coords = y_coords + float(args["add_y_coord"])
            # now, reconstruct output format strings
            out_text = ''
            for (i, x_coord) in enumerate(x_coords):
                out_text += "%f,%f " % (x_coords[i],y_coords[i])                
            element_coords.text = out_text               
        else:
            [x_coord, y_coord] = point_list[0].split(",")
            [x_coord, y_coord] = [float(x_coord), float(y_coord)]            
            if args.has_key("x_coord"):
                x_coord = float(args["x_coord"])
            if args.has_key("y_coord"):
                y_coord = float(args["y_coord"])
            if args.has_key("add_x_coord"):
                x_coord = x_coord + float(args["add_x_coord"])
            if args.has_key("add_y_coord"):
                y_coord = y_coord + float(args["add_y_coord"])
            element_coords.text = "%f,%f" % (x_coord, y_coord)
        return None

    def change_foliation_polarity(self, element):
        """change polarity of foliation element"""
        if element.get("NormalPolarity") == "true":
            element.set("NormalPolarity", "false")
        else:
            element.set("NormalPolarity", "true")
        
    def change_foliation(self, element, **args):
        """change foliation data, argument one or more of: azimuth, dip, 
        normalpolarity = true/false, x_coord, y_coord" or: add_dip, add_azimuth,
        add_x_coord, add_y_coord to add values to existing values!
        use if_name = and if_provenance = to add conditions!"""
        prov = element.get("Provenance")
        name = element.find("{"+self.xmlns+"}Data").get("Name")
        if args.has_key("if_name"):
            if args["if_name"] != name: return
        if args.has_key("if_provenance"):
            if args["if_provenance"] != prov: return
        element_fol = element.find("{"+self.xmlns+"}FoliationObservation")
        if args.has_key("dip"):
            element_fol.set("Dip", args["dip"])
        if args.has_key("azimuth"):
            element_fol.set("Azimuth", args["azimuth"])
        if args.has_key("nomalpolarity"):
            element_fol.set("NormalPolarity", args["normalpolarity"])
        #
        # To Do: logfile, if dip >90, azi > 360, ...
        #
        if args.has_key("add_dip"):
            dip_org = float(element_fol.get("Dip"))
            dip_new = dip_org + float(args["add_dip"])
            if dip_new > 90:
                dip_new = 180 - dip_new
                self.change_foliation_polarity(element_fol)
                azi_org = float(element_fol.get("Azimuth"))
                if azi_org < 180:
                    element_fol.set("Azimuth", str(azi_org+180))
                else:
                    element_fol.set("Azimuth", str(azi_org-180))
            element_fol.set("Dip", str(dip_new))
        if args.has_key("add_azimuth"):
            azi_org = float(element_fol.get("Azimuth"))
            azi_new = azi_org + float(args["add_azimuth"])
            if azi_new > 360.0: azi_new -= 360
            element_fol.set("Azimuth", str(azi_new))
        element_point = element_fol.find("{"+self.gml+"}Point")
        element_coords = element_point.find("{"+self.gml+"}coordinates")
        [x_coord, y_coord] = element_coords.text.split(",")
        [x_coord, y_coord] = [float(x_coord), float(y_coord)]
        if args.has_key("x_coord"):
            x_coord = float(args["x_coord"])
        if args.has_key("y_coord"):
            y_coord = float(args["y_coord"])
        if args.has_key("add_x_coord"):
            x_coord = x_coord + float(args["add_x_coord"])
        if args.has_key("add_y_coord"):
            y_coord = y_coord + float(args["add_y_coord"])
        element_coords.text = "%f,%f" % (x_coord, y_coord)
        return None
    
    def twt_to_depth(self, sec_element, formula, **args):
        """Convert all data within a section from twt to depth (including
        orientation data!!
        Input: section element with data points, conversion function as
        string with 't' as placeholder for twt-time, e.g. '2 * (-t) ** 2 + 18 * (-t)'
        ATTENTION: check if t negative
        optional arguments: 
        change_dip (boolean) : change dip angle in foliation data
        according to first derivative of twt-to-depth formula
        create_plot (boolean) : create summary plot with twt and converted depth
        for conversion formula control
        """            
        # Idea: stoachstic apporach to twt -> depth conversion: apply several
        # possible formulae for quality estimation of resulting model?
        struct_data = sec_element.find("{"+self.xmlns+"}Structural2DData")
        interfaces = struct_data.find("{"+self.xmlns+"}Interfaces").findall("{"+self.xmlns+"}Interface")
        # save data in list to create a plot to check validity of conversion
        t_list = []
        v_list = []
        for interface in interfaces:
            gml_coords_element = interface.find("{"+self.gml+"}LineString").find("{"+self.gml+"}coordinates")
            # check for correct decimal, column (cs) and text separator (ts)
            ts = gml_coords_element.get("ts")
            cs = gml_coords_element.get("cs")
            data_array = gml_coords_element.text.split(ts)
            # check if last entry is empty (seems to happen often), if so: delete!
            # print gml_coords_element.text
            if data_array[-1] == "": del data_array[-1]
            text_new = ""
            # apply conversion formula for z-direction, i.e. dv;
            # no change in x-direction -> du = 0 (but: maybe neccessary for specific situations?)
            for entry in data_array:
                parts = entry.split(cs)
                # get original values
                # t as varibale, as defined in formula (input)
                t = float(parts[1])
                v_new = eval(formula)
                du = 0
                text_new += "%f%s%f%s" % (float(parts[0])+du, cs, v_new, ts)
                # append to list for check-plot
                t_list.append(-t)
                v_list.append(v_new)
            # print text_new                                                            
            gml_coords_element.text = text_new
            # print gml_coords_element.text
        # now change foliation position and dip angle! (check: given as argument?)
        # for dip angle: numerical determination of first derivative for
        # twt to depth conversion formula?    
        if args.has_key("change_dip") and args["change_dip"]:
            print "change dip in seismic profile"
        
        # create check-plot
        if args.has_key("create_plot") and args["create_plot"]:
            print "Create plot with twt, converted depth pairs"            
            plot(t_list,v_list,'.', label = formula)
            title("TWT to depth: Converted data points\nSection: " + sec_element.get("Name"))
            xlabel("TWT [ms]")
            ylabel("Converted depth [m]")
            legend()
            grid(True)
            show()
            
    def get_pole_points(self, element):
        # function to plot data points in geomodeller element
        u = []
        v = []
        poles = element.getiterator("{"+self.xmlns+"}Pole-Weight")
        for pole in poles:
            u.append(pole.get("U"))
            v.append(pole.get("V"))
        
        return (u,v)
    
    def plot_points(self, element):
        # plot u,v points in simple 2D plot
        (u,v) = self.get_pole_points(element)
        plot(u,v,'o')
        name = element.get("Name")
        title(name)
        savefig(name+"_points.png")
            
    def write_xml(self, save_dir):
        """Write elementtree to xml-file""" 
        # to do: filename (and directory?) as optional argument"""
        # flo, 10/2008
        #file = "changed_model.xml"
        tree_new = ET.ElementTree(self.rootelement)
        print "Write tree to file " + save_dir
        tree_new.write(save_dir)
        #self.tree.write("changed_model.xml")
        self.tree.write(save_dir)
     
    def add_to_project_name(self, s):
        """add string s to project name, e.g. Number of uncertainty run"""
        name = self.rootelement.get("projectName")
        name_new = name + " " + s
        self.rootelement.set("projectName",name_new)
        
    def get_model_extent(self):
        """get extent of model
        returns (x_min, x_max, y_min, y_max, z_min, z_max)
        and saves extent in self.x_min, self.x_max, etc.
        """
        extent_parent = self.rootelement.find("{"+self.xmlns+"}Extent3DOfProject")
        extentbox3D = extent_parent.find("{"+self.xmlns+"}ExtentBox3D")
        extent3D = extentbox3D.find("{"+self.xmlns+"}Extent3D")
        extent_xy = extent3D.find("{"+self.xmlns+"}ExtentXY")
        extent_z = extent3D.find("{"+self.xmlns+"}ExtentZ")
        self.x_min = float(extent_xy.get("Xmin"))
        self.x_max = float(extent_xy.get("Xmax"))
        self.y_min = float(extent_xy.get("Ymin"))
        self.y_max = float(extent_xy.get("Ymax"))
        self.z_min = float(extent_z.get("Zmin"))
        self.z_max = float(extent_z.get("Zmax"))
        return (self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max)
        
    def get_model_range(self):
        """get model range from model extent, e.g. for automatic mesh generation"""
        (x_min, x_max, y_min, y_max, z_min, z_max) = self.get_model_extent()
        from numpy import abs
        self.range_x = abs(x_max - x_min)
        self.range_y = abs(y_max - y_min)
        self.range_z = abs(z_max - z_min)
        return (self.range_x, self.range_y, self.range_z)
    
    def create_TOUGH_formation_names(self, **kwds):
        """create formation names that are compatible with format required by TOUGH,
        i.e. String of length 5
        returns and stores as dictionary with Geomodeller Names as key and TOUGH names as entry
        (self.tough_formation_names)
        simply cuts original formation name to a name length of 5;
        if cut name already exists: create new name, three first string values followed by two integers
        that are subsequently increased
        optional keywods:
        out = string : set out formation to this name (for TOUGH2: no leading spaces allowed! set to 5 chars!)
        """
        # import str
        self.tough_formation_names = {}
        # check if self.formation_names list already exists, if not: create
        try: self.formation_names
        except AttributeError:
            self.formation_names = self.get_stratigraphy_list(**kwds)
        # create list with tough names to check if name already exists
        tough_name_list = []
        for i,name in enumerate(self.formation_names):
#            if self.formation_names[i] == '  out' or self.formation_names[i] == '  OUT':
#                tough_name_list.append("OUT  ")
#                continue
            cut_name = self.formation_names[i][0:5]
            if cut_name in tough_name_list:
                for j in range(100):
                    if "%3s%02d" % (cut_name[0:3],j) in tough_name_list:
                        continue
                    else: 
                        cut_name = "%3s%02d" % (cut_name[0:3],j)
                        tough_name_list.append(cut_name)
                        break
            else:
                tough_name_list.append(cut_name)        
            self.tough_formation_names[name] = "%5s" % str.upper(cut_name)
        return self.tough_formation_names
    

if __name__ == '__main__':
    print "main"   



'''Module with classes and methods to analyse and process exported geomodel grids

Created on 21/03/2014

@author: Florian Wellmann (some parts originally developed by Erik Schaeffer)
'''
import numpy as np
import matplotlib.pyplot as plt
# import mpl_toolkits
# from matplotlib.ticker import MultipleLocator, FormatStrFormatter

# to convert python variable types to cpp types
import ctypes
# to create array
from numpy.ctypeslib import ndpointer
# to create folder
import os
# read out and change xml file (here only used to read out model boundary information)
import geomodeller_xml_obj as GO


class GeoGrid():
    """Object definition for exported geomodel grids"""
    
    def __init__(self, **kwds):
        """GeoGrid contains methods to load, analyse, and process exported geomodel grids
        
        **Optional Keywords**:
            - *grid_filename* = string : filename of exported grid
            - *delxyz_filename* = string : file with model discretisation
            - *dimensions_filename* = string : file with model dimension (coordinates)
        """
        
        if kwds.has_key('grid_filename'):
            self.grid_filename = kwds['grid_filename']
        if kwds.has_key('delxyz_filename'):
            self.delxyz_filename = kwds['delxyz_filename']
        if kwds.has_key('dimensions_filename'):
            self.dimensions_filename = kwds['dimensions_filename']
            
    def __add__(self, G_other):
        """Combine grid with another GeoGrid if regions are overlapping"""
        # check overlap
        print self.ymin, self.ymax
        print G_other.ymin, G_other.ymax
        if (G_other.ymin < self.ymax and G_other.ymin > self.ymin):
            print("Grids overlapping in y-direction between %.0f and %.0f" %
                  (G_other.ymin, self.ymax))
    
    def load_grid(self):
        """Load exported grid, discretisation and dimensions from file"""
        if not hasattr(self, 'grid_filename'):
            raise AttributeError("Grid filename is not defined!")
        self.grid = np.loadtxt(self.grid_filename, 
                               delimiter = ',', 
                               dtype='int',
                               unpack=False)
        if hasattr(self, 'delxyz_filename'):
            self.load_delxyz(self.delxyz_filename)
            self.adjust_gridshape()
        if hasattr(self, 'dimensions_filename'):
            self.load_dimensions(self.dimensions_filename)    
    
    def load_delxyz(self, delxyz_filename):
        """Load grid discretisation from file"""
        del_lines = open(delxyz_filename, 'r').readlines()
        d0 = del_lines[0].split("*")
        self.delx = np.array([float(d0[1]) for _ in range(int(d0[0]))])
        d1 = del_lines[1].split("*")
        self.dely = np.array([float(d1[1]) for _ in range(int(d1[0]))])
        d2 = del_lines[2].split(",")[:-1]
        self.delz = np.array([float(d) for d in d2])
        (self.nx, self.ny, self.nz) = (len(self.delx), len(self.dely), len(self.delz))
        (self.extent_x, self.extent_y, self.extent_z) = (sum(self.delx), sum(self.dely), sum(self.delz))
        
    def set_delxyz(self, delxyz):
        """Set delx, dely, delz arrays explicitly and update additional attributes
        
        **Arguments**:
            - *delxyz* = (delx-array, dely-array, delz-array): arrays with cell dimensions
        """
        self.delx, self.dely, self.delz = delxyz
        (self.nx, self.ny, self.nz) = (len(self.delx), len(self.dely), len(self.delz))
        (self.extent_x, self.extent_y, self.extent_z) = (sum(self.delx), sum(self.dely), sum(self.delz))        
        
        
    def load_dimensions(self, dimensions_filename):
        """Load project dimensions from file"""
        dim = [float(d) for d in open(dimensions_filename, 'r').readlines()[1].split(",")]
        (self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax) = dim
        # calculate cell centre positions in real world coordinates
        
    def create_regular_grid(self, nx, ny, nz):
        """Define a regular grid from defined project boundaries and given discretisations"""
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.delx = [(self.xmax - self.xmin) / nx] * nx
        self.dely = [(self.ymax - self.ymin) / ny] * ny
        self.delz = [(self.zmax - self.zmin) / nz] * nz
        
    def get_dimensions_from_geomodeller(self, xml_filename):
        """Get grid dimensions from Geomodeller project
        
        **Arguments**:
            - *xml_filename* = string: filename of Geomodeller XML file
        """
        # Note: this implementation is based on the Geomodeller API
        # The boundaries could theoretically also be extracted from the XML file
        # directly, e.g. using the geomodeller_xml_obj module - but this would
        # require an additional module being loaded, so avoid here!
        filename_ctypes = ctypes.c_char_p(xml_filename)
        # get model boundaries
        lib = ctypes.CDLL('./libgeomod.so') #linux
        #lib = ctypes.windll.libgeomod #windows
        lib.get_model_bounds.restype = ndpointer(dtype=ctypes.c_int, shape=(6,))
        boundaries = lib.get_model_bounds(filename_ctypes)
        (self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax) = boundaries
        
    def update_from_geomodeller_project(self, xml_filename):
        """Update grid properties directly from Geomodeller project
        
        **Arguments**:
            - *xml_filename* = string: filename of Geomodeller XML file
        """
        # initialise grid
        self.grid = np.ndarray((self.nx, self.ny, self.nz), dtype="int")
        filename_ctypes = ctypes.c_char_p(xml_filename)
        # create cell position list with [x0, y0, z0, ... xn, yn, zn]
        cell_position = []
        ids = []
        for k in range(self.nz):
            for j in range(self.ny):
                for i in range(self.nx):
                    cell_position.append(self.cell_centers_x[i])
                    cell_position.append(self.cell_centers_y[j])
                    cell_position.append(self.cell_centers_z[k])
                    ids.append((i,j,k))
        
        # prepare variables for cpp function
        coord_ctypes = (ctypes.c_double * len(cell_position))(*cell_position)
        coord_len = len(cell_position)
        # call cpp function
        lib = ctypes.CDLL('./libgeomod.so')
        lib.compute_irregular_grid.restype = ndpointer(dtype=ctypes.c_int, shape=(coord_len/3,))
        formations_raw = lib.compute_irregular_grid(filename_ctypes, coord_ctypes, coord_len)
        # re-sort formations into array
        for i in range(len(formations_raw)):
            self.grid[ids[i][0],ids[i][1],ids[i][2]] = formations_raw[i]
            
        
    def set_dimensions(self, **kwds):
        """Set model dimensions, if no argument provided: xmin = 0, max = sum(delx) and accordingly for y,z
        
        **Optional keywords**:
            - *dim* = (xmin, xmax, ymin, ymax, zmin, zmax) : set dimensions explicitly
        """
        if kwds.has_key("dim"):
            (self.xmin, self.xmax, self.ymin, self.ymax, self.zmin, self.zmax) = kwds['dim']
        else:
            self.xmin, self.ymin, self.zmin = (0., 0., 0.)
            self.xmax, self.ymax, self.zmax = (sum(self.delx), sum(self.dely), sum(self.delz))
    
    def determine_cell_centers(self):
        """Determine cell centers for all coordinate directions in "real-world" coordinates"""
        if not hasattr(self, 'xmin'):
            raise AttributeError("Please define grid dimensions first")
        sum_delx = np.cumsum(self.delx)
        sum_dely = np.cumsum(self.dely)
        sum_delz = np.cumsum(self.delz)
        self.cell_centers_x = np.array([sum_delx[i] - self.delx[i] / 2. for i in range(self.nx)]) + self.xmin
        self.cell_centers_y = np.array([sum_dely[i] - self.dely[i] / 2. for i in range(self.ny)]) + self.ymin
        self.cell_centers_z = np.array([sum_delz[i] - self.delz[i] / 2. for i in range(self.nz)]) + self.zmin
        
    def adjust_gridshape(self):
        """Reshape numpy array to reflect model dimensions"""
        self.grid = np.reshape(self.grid, (self.nz, self.ny, self.nx))
        self.grid = np.swapaxes(self.grid, 0, 2)
        # self.grid = np.swapaxes(self.grid, 0, 1)
        
    def plot_section(self, direction, cell_pos='center', **kwds):
        """Plot a section through the model in a given coordinate direction
        
        **Arguments**:
            - *direction* = 'x', 'y', 'z' : coordinate direction for section position
            - *cell_pos* = int/'center','min','max' : cell position, can be given as
            value of cell id, or as 'center' (default), 'min', 'max' for simplicity
            
        **Optional Keywords**:
            - *cmap* = mpl.colormap : define colormap for plot (default: jet)
            - *colorbar* = bool: attach colorbar (default: True)
            - *rescale* = bool: rescale color bar to range of visible slice (default: False)
            - *ve* = float : vertical exageration (for plots in x,y-direction)
            - *figsize* = (x,y) : figsize settings for plot
            - *ax* = matplotlib.axis : add plot to this axis (default: new axis)
                    if axis is defined, the axis is returned and the plot not shown
                    Note: if ax is passed, colorbar is False per default!
            - *savefig* = bool : save figure to file (default: show)
            - *fig_filename* = string : filename to save figure
        """
        cmap = kwds.get('cmap', 'jet')
        rescale = kwds.get('rescale', False)
        ve = kwds.get('ve', 1.)
        figsize = kwds.get('figsize', (8,4))
        if direction == 'x':
            if type(cell_pos) == str:
                # decipher cell position
                if cell_pos == 'center' or cell_pos == 'centre':
                    pos = self.nx / 2
                elif cell_pos == 'min':
                    pos = 0
                elif cell_pos == 'max':
                    pos = self.nx
            else:
                pos = cell_pos
            grid_slice = self.grid[pos,:,:]
            grid_slice = grid_slice.transpose()
            aspect = self.extent_z/self.extent_x * ve
        elif direction == 'y':           
            if type(cell_pos) == str:
                # decipher cell position
                if cell_pos == 'center' or cell_pos == 'centre':
                    pos = self.ny / 2
                elif cell_pos == 'min':
                    pos = 0
                elif cell_pos == 'max':
                    pos = self.ny
            else:
                pos = cell_pos
            grid_slice = self.grid[:,pos,:]
            grid_slice = grid_slice.transpose()
            aspect = self.extent_z/self.extent_y * ve
        elif direction == 'z' :
            if type(cell_pos) == str:
                # decipher cell position
                if cell_pos == 'center' or cell_pos == 'centre':
                    pos = self.nz / 2
                elif cell_pos == 'min':
                    pos = 0
                elif cell_pos == 'max':
                    pos = self.nz
            else:
                pos = cell_pos
            grid_slice = self.grid[:,:,pos].transpose()           
            aspect = 1.
            
        if not kwds.has_key('ax'):
            colorbar = kwds.get('colorbar', True)
            # create new axis for plot
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
        else:
            colorbar = False
            ax = kwds['ax']
        
        if not hasattr(self, 'unit_ids'):
            self.determine_geology_ids()
        if rescale:
            vmin = np.min(grid_slice)
            vmax = np.max(grid_slice)
        else: # use global range for better comparison
            vmin = min(self.unit_ids)
            vmax = max(self.unit_ids)
        im = ax.imshow(grid_slice, interpolation='nearest',
                       cmap = cmap,
                       origin='lower_left',
                       vmin = vmin,
                       vmax = vmax,
                       aspect = aspect)
        if colorbar:
#            divider = mpl_toolkits.axes_grid1.make_axes_locatable(ax)
#            cax = divider.append_axes("bottom", size="5%", pad=0.2)
            cbar1 = fig.colorbar(im, orientation="horizontal")        
            ticks = np.arange(vmin, vmax+0.1, int(np.log2(vmax-vmin)/1.2), dtype='int')
            cbar1.set_ticks(ticks)
    #         cbar1.set_ticks(self.unit_ids[::int(np.log2(len(self.unit_ids)/2))])
            cbar1.set_label("Geology ID")
    #         cax.xaxis.set_major_formatter(FormatStrFormatter("%d"))
        
        
        if kwds.has_key("ax"):
            # return image and do not show
            return im
        
        if kwds.has_key('savefig') and kwds['savefig']:
            # save to file
            filename = kwds.get("fig_filename", "grid_section_direction_%s_pos_%d.png" %
                                (direction, cell_pos))
            plt.savefig(filename)
            
        else:
            plt.show()
        
    def export_to_vtk(self, vtk_filename="geo_grid", real_coords = True, **kwds):
        """Export grid to VTK for visualisation
        
        **Arguments**:
            - *vtk_filename* = string : vtk filename (obviously...)
            - *real_coords* = bool : model extent in "real world" coordinates
            
        **Optional Keywords**:
            - *grid* = numpy grid : grid to save to vtk (default: self.grid)
            - *var_name* = string : name of variable to plot (default: Geology)
        
        Note: requires pyevtk, available at: https://bitbucket.org/pauloh/pyevtk
        """
        grid = kwds.get("grid", self.grid)
        var_name = kwds.get("var_name", "Geology")
        from evtk.hl import gridToVTK
        # define coordinates
        x = np.zeros(self.nx + 1)
        y = np.zeros(self.ny + 1)
        z = np.zeros(self.nz + 1)
        x[1:] = np.cumsum(self.delx)
        y[1:] = np.cumsum(self.dely)
        z[1:] = np.cumsum(self.delz)
        
        # plot in coordinates
        if real_coords:
            x += self.xmin
            y += self.ymin
            z += self.zmin
        
        
        gridToVTK(vtk_filename, x, y, z,
                  cellData = {var_name: grid})
        
    def determine_geology_ids(self):
        """Determine all ids assigned to cells in the grid"""
        self.unit_ids = np.unique(self.grid)
        
    def get_name_mapping_from_file(self, filename):
        """Get the mapping between unit_ids in the model and real geological names
        from a csv file (e.g. the SHEMAT property file)
        
        **Arguments**:
            - *filename* = string : filename of csv file with id, name entries
        """
        self.unit_name = {}
        filelines = open(filename, 'r').readlines()[1:]
        for line in filelines:
            l = line.split(",")
            self.unit_name[int(l[1])] = l[0]
            
    def get_name_mapping_from_dict(self, unit_name_dict):
        """Get the name mapping directly from a dictionary
        
        **Arguments**:
            - *unit_name_dict* = dict with "name" : unit_id (int) pairs
        """
        self.unit_name = unit_name_dict
            
            
    def remap_ids(self, mapping_dictionary):
        """Remap geological unit ids to new ids as defined in mapping dictionary
        
        **Arguments**:
            - *mapping_dictionary* = dict : {1 : 1, 2 : 3, ...} : e.g.: retain
            id 1, but map id 2 to 3 (note: if id not specified, it will be retained)
        """
        # first step: create a single mesh for each id to avoid accidential
        # overwriting below (there might be a better solution...)
        if not hasattr(self, 'unit_ids'):
            self.determine_geology_ids()
        geol_grid_ind = {}
        for k,v in mapping_dictionary.items():
            geol_grid_ind[k] = self.grid == k
            print("Remap id %d -> %d" % (k,v))
        # now reassign values in actual grid
        for k,v in mapping_dictionary.items():
            print("Reassign id %d to grid" % v)
            self.grid[geol_grid_ind[k]] = v
        # update global geology ids
        self.determine_geology_ids()
        
    def determine_cell_volumes(self):
        """Determine cell volumes for each cell (e.g. for total formation volume calculation)"""
        self.cell_volume = np.ndarray(np.shape(self.grid))
        for k,dz in enumerate(self.delz):
            for j,dy in enumerate(self.dely):
                for i,dx in enumerate(self.delx):
                    self.cell_volume[i,j,k] = dx * dy * dz        
    
    
    def determine_indicator_grids(self):
        """Determine indicator grids for all geological units"""
        self.indicator_grids = {}
        if not hasattr(self, 'unit_ids'):
            self.determine_geology_ids()
        grid_ones = np.ones(np.shape(self.grid))
        for unit_id in self.unit_ids:
            self.indicator_grids[unit_id] = grid_ones * (self.grid == unit_id)
     
    def determine_id_volumes(self):
        """Determine the total volume of each unit id in the grid
        
        (for example for cell discretisation studies, etc."""
        if not hasattr(self, 'cell_volume'):
            self.determine_cell_volumes()
        if not hasattr(self, 'indicator_grids'):
            self.determine_indicator_grids()
        self.id_volumes = {}
        for unit_id in self.unit_ids:
            self.id_volumes[unit_id] = np.sum(self.indicator_grids[unit_id] * self.cell_volume)
        
    def print_unit_names_volumes(self):
        """Formatted output to STDOUT of unit names (or ids, if names are note
        defined) and calculated volumes
        """
        if not hasattr(self, 'id_vikumes'):
            self.determine_id_volumes()

        if hasattr(self, "unit_name"):
            # print with real geological names
            print("Total volumes of modelled geological units:\n")
            for unit_id in self.unit_ids:
                print("%26s : %.2f km^3" % (self.unit_name[unit_id], 
                                            self.id_volumes[unit_id]/1E9))
        else:
            # print with unit ids only
            print("Total volumes of modelled geological units:\n")
            for unit_id in self.unit_ids:
                print("%3d : %.2f km^3" % (unit_id, 
                                            self.id_volumes[unit_id]/1E9))
            
                
            
        
        
        
        
# ******************************************************************************                                                                               
#  Some additional helper functions   
# ******************************************************************************                                                                               
        
def combine_grids(G1, G2, direction, merge_type = 'keep_first', **kwds):
    """Combine two grids along one axis
    
    ..Note: this implementation assumes (for now) that the overlap is perfectly matching,
    i.e. grid cell sizes identical and at equal positions, or that they are perfectly adjacent!
    
    **Arguments**:
        - G1, G2 = GeoGrid : grids to be combined
        - direction = 'x', 'y', 'z': direction in which grids are combined
        - merge_type = method to combine grid:
            'keep_first' : keep elements of first grid (default)
            'keep_second' : keep elements of second grid
            'random' : randomly choose an element to retain
        ..Note: all other dimensions must be matching perfectly!!
    
    **Optional keywords**:
        - *overlap_analysis* = bool : perform a detailed analysis of the overlapping area, including
        mismatch. Also returns a second item, a GeoGrid with information on mismatch!
    
    **Returns**:
        - *G_comb* = GeoGrid with combined grid
        - *G_overlap* = Geogrid with analysis of overlap (of overlap_analysis=True)
    
    """
    overlap_analysis = kwds.get("overlap_analysis", False)
    # first step: determine overlap
    if direction == 'x':
        if G2.xmax > G1.xmax:
            overlap_min = G2.xmin
            overlap_max = G1.xmax
            # identifier alias for grids with higher/ lower values
            G_high = G2
            G_low = G1
        else:
            overlap_min = G1.xmin
            overlap_max = G2.xmax
            # identifier alias for grids with higher/ lower values
            G_high = G1
            G_low = G2

        # check if all other dimensions are perfectly matching 
        if (G1.ymin != G2.ymin) or (G1.zmin != G2.zmin) or \
            (G1.ymax != G2.ymax) or (G1.zmax != G2.zmax):
            raise ValueError("Other dimensions (apart from %s) not perfectly matching! Check and try again!" % direction)    
    
    elif direction == 'y':
        if G2.ymax > G1.ymax:
            overlap_min = G2.ymin
            overlap_max = G1.ymax
            # identifier alias for grids with higher/ lower values
            G_high = G2
            G_low = G1
        else:
            overlap_min = G1.ymin
            overlap_max = G2.ymax
            # identifier alias for grids with higher/ lower values
            G_high = G1
            G_low = G2

        # check if all other dimensions are perfectly matching 
        if (G1.xmin != G2.xmin) or (G1.zmin != G2.zmin) or \
            (G1.xmax != G2.xmax) or (G1.zmax != G2.zmax):
            raise ValueError("Other dimensions (apart from %s) not perfectly matching! Check and try again!" % direction)

    elif direction == 'z':
        if G2.zmax > G1.zmax:
            overlap_min = G2.zmin
            overlap_max = G1.zmax
            # identifier alias for grids with higher/ lower values
            G_high = G2
            G_low = G1
        else:
            overlap_min = G1.zmin
            overlap_max = G2.zmax
            # identifier alias for grids with higher/ lower values
            G_high = G1
            G_low = G2

        # check if all other dimensions are perfectly matching 
        if (G1.ymin != G2.ymin) or (G1.xmin != G2.xmin) or \
            (G1.ymax != G2.ymax) or (G1.xmax != G2.xmax):
            raise ValueError("Other dimensions (apart from %s) not perfectly matching! Check and try again!" % direction)

    overlap = overlap_max - overlap_min
        
    if overlap == 0:
        print("Grids perfectly adjacent")
    elif overlap < 0:
        raise ValueError("No overlap between grids! Check and try again!")
    else:
        print("Positive overlap in %s direction of %f meters" % (direction, overlap))
      
    # determine cell centers
    G1.determine_cell_centers()
    G2.determine_cell_centers()
    
    # intialise new grid
    G_comb = GeoGrid()
    # initialise overlap grid, if analyis performed
    if overlap_analysis:
        G_overlap = GeoGrid()
    
    
    if direction == 'x':
        pass
    elif direction == 'y':
        #=======================================================================
        # Perform overlap analysis
        #=======================================================================
        
        # initialise overlap grid with dimensions of overlap
        G_overlap.set_dimensions(dim = (G1.xmin, G1.xmax, overlap_min, overlap_max, G1.zmin, G1.zmax))
        G_low_ids = np.where(G_low.cell_centers_y > overlap_min)[0]
        G_high_ids = np.where(G_high.cell_centers_y < overlap_max)[0]
        delx = G1.delx
        dely = G_low.dely[G_low_ids]
        delz = G1.delz
        G_overlap.set_delxyz((delx, dely, delz))
        # check if overlap region is identical
        if not (len(G_low_ids) == len(G_high_ids)):
            raise ValueError("Overlap length not identical, please check and try again!")
        # now: determine overlap mismatch
        G_overlap.grid = G_low.grid[:,G_low_ids,:] - G_high.grid[:,G_high_ids,:]
        # for some very strange reason, this next step is necessary to enable the VTK
        # export with pyevtk - looks like a bug in pyevtk...
        G_overlap.grid = G_overlap.grid + np.zeros(G_overlap.grid.shape)
        #
        
        #=======================================================================
        # Set up combined grid
        #=======================================================================
        
        G_comb.set_dimensions(dim = (G1.xmin, G1.xmax, G_low.ymin, G_high.ymax, G1.zmin, G1.zmax))
        # combine dely arrays
        dely = np.hstack((G_low.dely[:G_low_ids[0]], G_high.dely))
        G_comb.set_delxyz((delx, dely, delz))        
        
        #=======================================================================
        # Now merge grids
        #=======================================================================
        if merge_type == 'keep_first':
            if G1.ymax > G2.ymax:
                G_comb.grid = np.concatenate((G2.grid[:,:G_low_ids[0],:], G1.grid), axis=1)
            else:
                G_comb.grid = np.concatenate((G1.grid, G2.grid[:,:G_low_ids[0],:]), axis=1)
        
        elif merge_type == 'keep_second':
            pass
        elif merge_type == 'random':
            pass
        else:
            raise ValueError("Merge type %s not recognised! Please check and try again!" % merge_type)
        
    elif direction == 'z':
        pass
      
    # Return combined grid and results of overlap analysis, if determined  
    if overlap_analysis:
        return G_comb, G_overlap
    else:
        return G_comb
    
    

def regular_geogrid_from_geomodeller(xml_filename, nx, ny, nz):
    """Create a regular geogrid from a geomodeller project

    **Arguments**:
        - *xml_filename* = string : Geomodeller porject file
        - *nx*, *ny*, *nz* = int : number of cells in each direction
    """
    G1 = GeoGrid()
    G1.get_dimensions_from_geomodeller(xml_filename)
    G1.create_regular_grid(nx, ny, nz)
    G1.determine_cell_centers()
    G1.update_from_geomodeller_project(xml_filename)        
    return G1

if __name__ == '__main__':
    pass

























import os
from distutils.core import setup

# Advise from https://pythonhosted.org/an_example_pypi_project/setuptools.html
# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='pygeomod',
      version='0.1',
      description='Python wrappers for Geomodeller API and input files',
      long_description=read('README.md'),
      classifiers=[
          "Development Status :: 2 - Pre-Alpha",
          "Intended Audience :: Science/Research",
          "Programming Language :: Python :: 2.7",
          "Intended Audience :: Education",
          "Topic :: Education",
          "Topic :: Scientific/Engineering",
          "Topic :: Scientific/Engineering :: Physics",
          "Topic :: Scientific/Engineering :: Visualization",
          "License :: OSI Approved :: MIT License",
      ],
      url='http://github.com/flohorovicic/pygeomod',
      author='Florian Wellmann',
      author_email='flohorovicic@gmail.com',
      license='MIT',
      packages=['pygeomod'],
      # package_dir = {'output' : 'pynoddy', 'history' : 'pynoddy'},
      py_modules = ['pygeomod.geomodeller_xml_obj', 'pygeomod.geogrid'])
